#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" script for running in the terminal
@author: YangyangLi
@contact:li002252@umn.edu
@license: MIT Licence
@file: __main__.py
@time: 2020/12/21 5:02 PM
"""
from cli import cli

if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" class for scanning cryptic exons based on previous _result ::
Junction detector and junction annotator
"""
from typing import Iterable

import numpy as np
import pandas as pd

from .utils import timethis


def check(axis) -> np.array:
    """check if cryptic exon has children

    Which means that check if the cryptic exon is split by others junction reads
    in terms of start and end position

    :param axis: an array, a junction read, included start and end
    :type axis: numpy.array
    :return: whether junction read can split cryptic exon
    :rtype: bool
    """

    return (axis[0] < 0) & (axis[1] > 0)


def assign_value(df_ce, ces, ns, ce_id, ns_id) -> None:
    """assign value of `child column` for every cryptic exons that contains junction reads with N type

    Given the start and end of cryptic exons and junction reads,
    Note: types of junction reads contains N, D, A, DA, NDA. For details:
    https://regtools.readthedocs.io/en/latest/commands/junctions-annotate/

    :param df_ce: pandas.DataFrame of cryptic exons
    :type df_ce: pandas.DataFrame
    :param ces: cryptic exons' pandas.DataFrame,
        which has gene ids that both cryptic exon and junction
        read own. It only contain gene id, start and end
    :type ces: pandas.DataFrame
    :param ns: pandas.DataFrame of junction reads whose type is N,
        as well as has same gene id as cryptic exons
    :type ns: pandas.DataFrame
    :param ce_id: index of gene id of cryptic exons which has junction reads as children
    :type ce_id: numpy.array
    :param ns_id: index of gene id of junction reads with N type, which have junction reads as children
    :type ns_id: numpy.array
    """
    children_info = f"{ns.iloc[ns_id, 0]}-{ns.iloc[ns_id, 1]},"  # start-end
    gene_id = ces.index[ce_id]
    length = ces.iloc[ce_id, 1] - ces.iloc[ce_id, 0]
    df_ce.loc[(gene_id, length), "children"] += children_info


def split_ce(df_ce, df_n) -> Iterable:
    """Iterator: check whether detected cryptic exons are split by other junction reads

    :param df_ce: pandas.DataFrame of cryptic exons return from :func:`find_ce`
    :type df_ce: pandas.DataFrame
    :param df_n: pandas.DataFrame of junction reads with N type
    :type df_n: pandas.DataFrame
    :return: param:`df_ce` with new column `children`
    :rtype: iterator
    """
    # set gene as index
    df_ce.set_index("gene", inplace=True)
    df_n.set_index("gene", inplace=True)
    # get genes contained in both cryptic exons and junction reads with N type
    gene_ind = set(df_ce.index) & set(df_n.index)

    ces = df_ce.loc[gene_ind, ["start", "end"]]
    ns = df_n.loc[gene_ind, ["start", "end"]]
    ces_values = ces.values
    ns_values = ns.values
    # set another index in case that one gene has more than one cryptic exons
    df_ce.set_index("length", append=True, inplace=True)

    if ces_values.size > 1:

        x = np.apply_along_axis(
            lambda row: row - ns_values,
            1,
            ces_values,
        )  # iter for row

        bool_re = np.apply_along_axis(check, 2, x)

        position = np.argwhere(
            bool_re,
        )  # return row id, n_df id[[ce_id,n_id],[ce_id,n_id]]

        if position.size >= 1:

            for ce_id, ns_id in position:
                assign_value(df_ce, ces, ns, ce_id, ns_id)

    else:

        x = ces_values - ns_values
        bool_x = (x[:, 0] < 0) & (x[:, 1] > 0)
        position = np.argwhere(bool_x)  # return [[n_id],[n_id]]

        if position.size >= 1:
            for ns_id in position.flatten():
                assign_value(df_ce, ces, ns, ce_id=0, ns_id=ns_id)

    yield df_ce


def find_ce(groups) -> Iterable:
    """parse _result getting from annotations in order to detect cryptic exons

    :param groups:  annotated junction reads are grouped by `strand` and `type`
    :type groups: Groupby object return from ``Dataframe.groupby``
    :return: pd.DataFrame of two strands
    :rtype: Iterable
    """
    # old column  and new column
    pick_col = (
        "chrom",
        "end_D",
        "start_A",
        "start",
        "end",
        "strand",
        "score_DA",
        "score_D",
        "score",
        "gene",
    )
    rename_col = (
        "chrom",
        "start",
        "end",
        "start_DA",
        "end_DA",
        "strand",
        "score_DA",
        "score_D",
        "score_A",
        "gene",
    )

    for strand in ("+", "-"):
        da = groups.get_group((strand, "DA"))
        d = groups.get_group((strand, "D"))
        a = groups.get_group((strand, "A"))
        n = groups.get_group((strand, "N"))

        yield from (
            da.merge(
                d,
                left_on=["chrom", "start", "gene"],
                right_on=["chrom", "start", "gene"],
                suffixes=("_DA", "_D"),
            )
            .merge(
                a,
                left_on=["chrom", "end_DA", "gene"],
                right_on=["chrom", "end", "gene"],
                suffixes=("", "_A"),
            )[lambda df: df["end_D"] < df["start_A"]]
            .pipe(lambda df: df.loc[:, pick_col])
            .rename(columns=dict(zip(pick_col, rename_col)))
            .assign(length=lambda df: df.end - df.start)
            .assign(children="")
            .pipe(split_ce, n)
        )


class Scanner:
    """class for scanning cryptic exons based on annotated junction reads

    :param cutoff: cutoff used to filter junction reads with relatively low score or depth
    :type cutoff: int
    :param output: filename of _result
    :type output: str
    """

    def __init__(self, cutoff, output):
        """Constructor for Scanner"""
        self.cutoff = cutoff
        self.output = output
        self._result = None

    def __repr__(self):
        return f"Scanner({self.cutoff!r})"

    @timethis(name="File Writer for Scanner", message="FINISHED")
    def write2file(self):
        """ start iterator and write _result to file"""
        pd.concat(self._result).to_csv(self.output, sep="\t", encoding="utf8")

    @timethis(name="Cryptic Exon Scanner", message="FINISHED")
    def run(self, junctionmap) -> Iterable:
        """run program to scan cryptic exons

        :param junctionmap: instance from :class:`ce_detector.detector.JunctionMap`
        :type junctionmap: instance
        :return: temporary result used to store cryptic exons
        :rtype: Iterable
        """
        self._result = (
            pd.DataFrame(
                [
                    [read.chrom, read.start, read.end, read.strand, read.score, *info]
                    for read in junctionmap
                    for info in read.information
                    if read.score >= self.cutoff
                ],
                columns=[
                    "chrom",
                    "start",
                    "end",
                    "strand",
                    "score",
                    "type",
                    "dk",
                    "ak",
                    "gene",
                ],
            )
            .pipe(lambda df: df.groupby(["strand", "type"]))
            .pipe(find_ce)
        )

        return self._result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: YangyangLi
@contact:li002252@umn.edu
@version: 0.0.1
@license: MIT Licence
@file: cli.py.py
@time: 2020/12/28 10:21 PM
"""
import click
import gffutils
from rich.traceback import install

from . import __version__
from .annotator import Annotator
from .detector import JunctionDetector
from .scanner import Scanner

install()

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"], max_content_width=150)


@click.group(
    context_settings=CONTEXT_SETTINGS,
    options_metavar="<options>",
    subcommand_metavar="<command>",
)
@click.version_option(__version__)
def cli():
    """program designed for detecting cryptic exons

    Needed file types:

    \b
    1. bam file
    2. genome reference
    3. annotation file
    """
    pass


@cli.command(
    "build",
    short_help="build database for annotation file",
    options_metavar="<options>",
)
@click.argument("gff", type=click.Path(exists=True))
@click.option(
    "--out-directory",
    "-o",
    type=click.Path(exists=True),
    default=".",
    show_default=True,
    metavar="<path>",
)
def build(gff, out_directory):
    """build database for annotation file

    build database of annotation file in order to use the database to annotate junctions reads later.
    this process is time-confusing so that you may need to prepare a cup of coffee!
    file of database named {prefix of annotation file}.db

    \f
    :param gff: the path of annotation file
    :type gff: str
    :param out_directory: the path of _result of database
    :type out_directory: str
    :return: {out directory}/{prefix of annotation file}.db
    """
    _ = gffutils.create_db(
        gff,
        out_directory,
        merge_strategy="create_unique",
        keep_order=True,
    )


@cli.command("detect", short_help="scan cryptic exons", options_metavar="<options>")
@click.option(
    "--bam",
    "-b",
    help="The bam file (Bam or Sam format)",
    required=True,
    type=click.Path(exists=True),
    metavar="<path>",
)
@click.option(
    "--reference",
    "-r",
    help="The reference fasta (.fna) file, which contains index file(*.fai)",
    required=True,
    type=click.Path(exists=True),
    metavar="<path>",
)
@click.option(
    "--quality",
    "-q",
    help="The threshold to filter low quality reads",
    default=0,
    type=click.INT,
    show_default=True,
    metavar="<int>",
)
@click.option(
    "--out",
    "-o",
    help="The output file of detected cryptic exons",
    default="cryptic_exons.bed",
    type=click.STRING,
    show_default=True,
    metavar="<path>",
)
@click.option(
    "--gffdb",
    "-db",
    help="The database of annotation file",
    type=click.Path(exists=True),
    required=True,
    metavar="<path>",
)
@click.option(
    "--cutoff",
    "-c",
    help="cutoff for filtering junction reads wth low depth during scanning cryptic exons",
    type=click.INT,
    default=1,
    show_default=True,
    metavar="<int>",
)
@click.option(
    "--out-ann",
    "-oa",
    help="The output file of annotated junction reads",
    default="annotated_junctions.bed",
    type=click.File(mode="w", encoding="utf-8"),
    show_default=True,
    metavar="<path>",
)
def detect(bam, reference, quality, gffdb, cutoff, out, out_ann):
    """detect junction reads and scan cryptic exons

    \b
    analysis following steps below:
    1. detect junction reads in terms of bam file
    2. annotate junction reads in terms of genome reference and annotation file
    3. scan cryptic exons according to its definition

    \f
    :param out_ann:
    :type out_ann:
    :param cutoff:
    :type cutoff:
    :param bam: bam file
    :type bam: str
    :param reference: genome reference file
    :type reference: str
    :param quality: quality used to filter low quality reads.Default:0
    :type quality: int
    :param gffdb: database file of annotation file
    :type gffdb: str
    :param out: file name of detected cryptic exons
    :type out: str
    """

    # click.secho((f'\nProgram Start.\nParameters:\nReference: {reference}'
    #              f'\nQuality Threshold: {quality}\nOutput: {out}\n '), fg='red', bold=True)

    detector = JunctionDetector(
        bam,
        reference,
        quality,
    )
    junctionmap = detector.run()

    annotator = Annotator(junctionmap, gffdb)
    annotator.run()

    scanner = Scanner(cutoff=cutoff, output=out)
    scanner.run(annotator.junctionMap)
    scanner.write2file()


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# from click.testing import CliRunner
import sys

sys.path.append("..")
# from ce_detector import cli

#
# class TestCli:
#     def test_build(self):
#         runner = CliRunner()
#         with runner.isolated_filesystem():
#             with open('test_annotation.gff', 'w') as f:
#                 f.write('This is annotation file')
#
#             _result = runner.invoke(cli, [
#                 'build',
#                 'test_annotation.gff',
#             ])
#             assert _result.exit_code == 0
#             assert 'database' in _result.output
#
#     def test_detect(self):
#         runner = CliRunner()
#
#         with runner.isolated_filesystem():
#             with open('test_bam.bam', 'w') as f:
#                 f.write('This is bam file')
#             with open('test_reference.fa', 'w') as f:
#                 f.write('This is genome reference file')
#             with open('test_gffdb.db', 'w') as f:
#                 f.write('This is gffdb file')
#
#             _result = runner.invoke(cli, [
#                 'detect',
#                 '--bam', 'test_bam.bam',
#                 '--reference', 'test_reference.fa',
#                 '--gffdb', 'test_gffdb.db',
#                 '--quality', '30',
#                 '--out', 'test_out.bed'
#             ])
#
#             assert _result.exit_code == 0
#             assert 'test_bam.bam' in _result.output
#             assert 'test_reference.fa' in _result.output
#             assert 'test_gffdb.db' in _result.output
#             assert 'quality=30' in _result.output
#             assert 'test_out.bed' in _result.output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""class for annotating junction reads
"""
from __future__ import annotations

from collections import defaultdict
from typing import Any

import gffutils
import numpy as np

from .utils import get_yaml
from .utils import timethis

# HardCode the information of chromosome because its name of two ref are not identical
CHROMS = get_yaml()["chr2hg38"]


class Annotator:
    """annotate junction reads

    :param junctionmap:  instance return :class:`ce_detector.detector.JunctionMap`
    :type junctionmap: instance
    :param database: database of annotation files
    :type database: Any
    :param output: filename of annotated junction reads. Defaults to None
    :type output: TestIo
    """

    def __init__(self, junctionmap, database: Any, output=None):
        """Constructor of Annotator"""
        self.junctionMap = junctionmap
        self.database = gffutils.FeatureDB(database)
        self.output = output

    @staticmethod
    def detect_property(start, end, junction_list):
        """detect type of slice, number of skipped donors and number of skipped acceptors

        type of slice including D A DA N NDA

        :param start: start of junction read
        :type start: int
        :param end: end of junction read
        :type end: int
        :param junction_list: gene list of junction reads
        :type junction_list: numpy.array
        :return: type of slice, number of skipped donors, number of skipped acceptors
        """
        known_donors = junction_list[:, 0]
        known_acceptors = junction_list[:, 1]

        donors_skipped = ((start + 1 < known_donors) & (known_donors < end)).sum()

        acceptors_skipped = (
            (start + 1 < known_acceptors) & (known_acceptors < end)
        ).sum()

        if [start + 1, end] in junction_list.tolist():

            reads_type = "DA"

        elif (start + 1 in known_donors) and (end in known_acceptors):

            reads_type = "NDA"

        elif (start + 1 in known_donors) and (end not in known_acceptors):

            reads_type = "D"

        elif (start + 1 not in junction_list[:, 0]) and (end in junction_list[:, 1]):

            reads_type = "A"

        else:

            reads_type = "N"

        return reads_type, donors_skipped, acceptors_skipped

    def annotate_junction(self, read, result, db):
        """annotate junction reads and write results to file

        :param read: junction read return :class:`ce_detector.detector.Read`
        :type read: instance
        :param result: gene list used for annotation
        :type result: defaultdict[Any, Any]
        :param db: database of annotation file
        :type db: instance of file
        """
        chrom = read.chrom
        start, end = read.start, read.end
        region = f"{CHROMS[chrom]}:{start}-{end}"  # change chromosome

        gene_list = []

        for gene in db.features_of_type(("gene"), limit=region):

            gene_list.append(gene.id)
            # new gene
            if gene.id not in result:
                # transcript

                for transcript in db.children(
                    gene,
                    level=1,
                    featuretype=("primary_transcript", "transcript", "mRNA"),
                ):  # mRNA may drop out

                    # find all position of introns for every gene known junctions
                    for junction in db.interfeatures(
                        db.children(
                            transcript,
                            level=1,
                            featuretype="exon",
                            order_by="start",
                        ),
                        new_featuretype="intron",
                    ):
                        result[gene.id].append([junction.start, junction.end])

                if (
                    len(result[gene.id]) >= 1
                ):  # drop gene with only one exon in terms of number of intron

                    result[gene.id] = np.unique(result[gene.id], axis=0)

                else:
                    result.pop(gene.id)
                    gene_list.pop()

        # annotate junctions reads

        for gene in gene_list:
            junction_list = result[gene]

            reads_type, donors_skipped, acceptors_skipped = self.detect_property(
                start,
                end,
                junction_list,
            )
            #         read.type = reads_type
            #     output.write(f'{reads_information}\t{reads_type}\t{gene}\n')

            read.information.append(
                [reads_type, donors_skipped, acceptors_skipped, gene],
            )
            if self.output:
                self.output.write(
                    f"{read}\t{reads_type}\t{donors_skipped}\t{acceptors_skipped}\t{gene}\n",
                )

    @timethis(name="Junction Annotator", message="FINISHED")
    def run(self):
        """main function used to annotate junction reads

        pick all genes covered by one junction read and annotate all of them:
        type of slice, number of skipped donors and number of skipped acceptors

        """

        result = defaultdict(list)

        # logger.info(f'Begin to annotate junctions!')

        for _index, read in enumerate(self.junctionMap):
            #         if index < 10:
            self.annotate_junction(read, result, self.database)



# -*- coding: utf-8 -*-
"""Top-level package for ce-detector."""

__author__ = """YangyangLi"""
__email__ = "li002252@umn.edu "
__version__ = "__version__ = '0.1.12'"

# from .annotator import Annotator
# from .ce_scanner import Scanner
# from .detector import JunctionDetector
# from .cli import cli
# from . import utils



# -*- coding: utf-8 -*-
import logging
import time
from functools import partial
from functools import wraps
from os.path import dirname
from os.path import join

import importlib_resources
import yaml
from rich.logging import RichHandler


class Timer:
    """construct  Timer to show working time of tasks"""

    def __init__(self, func=time.perf_counter):
        """init values

        Args:
            func : Defaults to time.perf_counter.
        """
        self.elapsed = 0.0

        self._func = func

        self._start = None

    def start(self):
        """start a task

        Raises:
            RuntimeError:[if task has started then raise error]
        """

        if self._start is not None:
            raise RuntimeError("Already started")

        self._start = self._func()

    def stop(self):
        """end a task

        Raises:
            RuntimeError[if task has not started then raise error]
        """
        if self._start is None:
            raise RuntimeError("Not started")

        end = self._func()

        self.elapsed += end - self._start

        self._start = None

    def reset(self):
        """reset the working time"""
        self.elapsed = 0

    def running(self):
        """check if task is running

        Returns:
            bool
        """

        return self._start is not None

    def __enter__(self):
        """function used to address 'with text'"""

        self.start()

        return self

    def __exit__(self, *args):
        """function used to address 'with text'"""

        self.stop()


def rich_logger(logger_name, create_file=False, level=logging.INFO):
    """set logger and output console

    :param level: level for logging
    :param logger_name: logger name
    :type logger_name: str
    :param create_file: whether creat file to store log
    :type create_file: bool
    :return: logger
    :rtype: instance
    """
    # create logger for prd_ci
    log = logging.getLogger(logger_name)

    # create formatter and add it to the handlers
    template = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    formatter = logging.Formatter(template)

    logging.basicConfig(
        level=level,
        format=template,
        handlers=[
            RichHandler(markup=True, show_level=False, show_time=False),
        ],
    )

    if create_file:
        # create file handler for logger.
        fh = logging.FileHandler(f"{logger_name}_log.txt")

        fh.setLevel(level=logging.DEBUG)

        fh.setFormatter(formatter)

    # add handlers to logger.
    if create_file:
        log.addHandler(fh)

    return log


def get_logger(logger_name, create_file=False, level=logging.INFO):
    """set logger and output console

    :param level: level for logging
    :param logger_name: logger name
    :type logger_name: str
    :param create_file: whether creat file to store log
    :type create_file: bool
    :return: logger
    :rtype: instance
    """
    # create logger for prd_ci
    log = logging.getLogger(logger_name)

    log.setLevel(level=level)

    # create formatter and add it to the handlers
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    if create_file:
        # create file handler for logger.
        fh = logging.FileHandler(f"{logger_name}_log.txt")

        fh.setLevel(level=logging.DEBUG)

        fh.setFormatter(formatter)

    # relate console handler for logger.
    ch = logging.StreamHandler()
    ch.setLevel(level=logging.DEBUG)
    ch.setFormatter(formatter)

    # add handlers to logger.
    if create_file:
        log.addHandler(fh)

    log.addHandler(ch)

    return log


def get_yaml():
    """get information of chromosome stored in yaml file
    :return: chromosome values
    :rtype: dict
    """
    if __package__:
        path = join(importlib_resources.files(__package__).as_posix(), "chromosome.yml")
    else:
        path = join(dirname(__file__), "chromosome.yml")
    return yaml.safe_load(open(path))


def timethis(
    func=None,
    level=logging.INFO,
    name=None,
    message=None,
    creat_file=False,
    rich=True,
):
    if func is None:
        return partial(
            timethis,
            level=level,
            name=name,
            message=message,
            creat_file=creat_file,
        )

    logname = name if name else func.__module__
    log = (
        rich_logger(logname, create_file=creat_file)
        if rich
        else get_logger(logname, create_file=creat_file)
    )
    logmsg = message if message else func.__name__

    @wraps(func)
    def wrapper(*args, **kwargs):
        with Timer() as t:
            temp = func(*args, **kwargs)

        log.info(f"[bold green]{logmsg} CONSUMING {t.elapsed:.2f}s")
        return temp

    return wrapper



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""class for detecting junction reads

"""
import pysam as ps

from .utils import get_yaml
from .utils import timethis

# change keys to be consist with chromosome's values
CHROMS = get_yaml()["chr2hg38"]

POSITIVE_SITE, NIGATIVE_SITE = [("GT", "AG"), ("AT", "AC"), ("GC", "AG")], [
    ("CT", "AC"),
    ("GT", "AT"),
    ("CT", "GC"),
]


class Read:
    """build a read class for storing information of every junction read

    :param chrom: chromosome of genome
    :type chrom: str
    :param start: start position of junction read
    :type start: int
    :param end: end position of junction read
    :type end: int
    :param idn: index of junction read
    :type idn: int
    :param score: support of junction read
    :type score: int
    :param strand: direction of junction read (-|+)
    :type strand: str
    :param anchor: anchor of junction read
    :type anchor: str
    :param acceptor: acceptor of junction read
    :type acceptor: str

    """

    __slots__ = (
        "chrom",
        "start",
        "end",
        "idn",
        "score",
        "strand",
        "anchor",
        "acceptor",
        "information",
    )

    def __init__(self, chrom, start, end, idn, score, strand, anchor, acceptor):
        self.chrom, self.start, self.end = chrom, start, end
        self.idn, self.score, self.strand = idn, score, strand
        self.anchor, self.acceptor = anchor, acceptor
        self.information = []

    @property
    def identifiers(self):
        return f"{self.chrom}_{self.start}_{self.end}"

    def __repr__(self):
        return (
            f"Read({self.chrom!r}, {self.start!r}, {self.end!r}, {self.idn!r}, "
            f"{self.score!r}, {self.strand!r}, {self.anchor!r}, {self.acceptor!r})"
        )

    def __str__(self):
        return "\t".join(
            map(
                str,
                [
                    self.chrom,
                    self.start,
                    self.end,
                    self.idn,
                    self.score,
                    self.strand,
                    f"{self.anchor}-{self.acceptor}",
                ],
            ),
        )


class JunctionMap:
    """build a class to store information of all junction reads"""

    def __init__(self):
        self.junctionList = {}

    def add_read(self, read):
        """add read to  junctionlist

        :param read: instance from Read
        :type read: instance
        """
        self.junctionList[read.identifiers] = read

    def get_read(self, identifiers):
        """get read from junctionlist according to identifiers

        :param identifiers: identifier for every read: chrom_start_end
        :type identifiers: int
        :return: instance from Read
        :rtype: instance
        """
        return self.junctionList[identifiers]

    def __contains__(self, identifiers):
        """check if read is in junctionlist according to identifiers

        :param identifiers: identifier for every read: chrom_start_end
        :type identifiers: str
        :return: whether Read is in JunctionMap
        :rtype: bool
        """
        return identifiers in self.junctionList

    def __iter__(self):
        """iterate every read in junctionlist"""
        return iter(self.junctionList.values())

    def write2file(self, output, header=None):
        """write all reads in junctionlist to file

        :param output: file name of output
        :type output: str or TextIo
        :param header: header of output
        :type header: str
        """
        try:
            output.write(f"{header}\n")
        except AttributeError:
            output = open(output, "w")
            output.write(f"{header}\n")
        finally:
            for read in self:
                output.write(f"{read}\n")
            output.close()


class JunctionDetector:
    """class for detecting junction reads and record position

    :param bam_file: bam file
    :type bam_file: str
    :param output: filename of output
    :type output: str
    :param reference: filename of genome reference
    :type reference: str
    :param quality: quality for filtering junction reads
    :type quality: int
    """

    def __init__(self, bam_file, reference, quality=0, output=None):

        self.bam_file, self.output = bam_file, output

        self.reference = reference

        self.quality = quality

    @staticmethod
    def check_strand(anchor, acceptor):
        """check type of strand

        :param anchor: anchor of read
        :type anchor: str
        :param acceptor: acceptor of read
        :type acceptor: str
        :return: type of strand (-|+)
        :rtype: str
        """
        strand = "N"

        if (anchor, acceptor) in POSITIVE_SITE:

            strand = "+"

        elif (anchor, acceptor) in NIGATIVE_SITE:

            strand = "-"

        return strand

    def worker(self, bam_file, reference, chrom, quality, idn, junctionmap):
        """find junction reads and annotate slice site

        :param bam_file: handle of bam_file
        :type bam_file: instance
        :param reference: handle of reference
        :type reference: instance
        :param chrom: chromosome
        :type chrom: str
        :param quality: quality for filtering reads
        :type quality: int
        :param idn: identifier of reads
        :type idn: int
        :param junctionmap: instance from junctionmap
        :type junctionmap: instance
        :return: instance from junctionmap
        :rtype: instance
        """
        # detect junction reads
        junction_regions = bam_file.find_introns(
            [
                r
                for r in bam_file.fetch(contig=chrom)  # chrom
                if r.mapping_quality > quality
            ],
        )

        # annotate slice sites
        for ((start, end), score) in junction_regions.items():
            junction_bases = reference.fetch(
                reference=CHROMS[chrom],
                start=start,
                end=end,  # change chrom
            )
            anchor, acceptor = junction_bases[:2].upper(), junction_bases[-2:].upper()

            strand = self.check_strand(anchor, acceptor)
            read = Read(chrom, start, end, idn + 1, score, strand, anchor, acceptor)
            junctionmap.add_read(read)
            #             line = f'{chrom}\t{start}\t{end}\t{idn+1}\t{score}\t{strand}\t{anchor}-{acceptor}'

            idn += 1

    @timethis(name="Junction detector", message="FINISHED")
    def run(self):
        """detect junction reads and annotate slice site, write results to file

        :return: instance from junctionmap
        :rtype: instance
        """

        idn = 0

        reference = ps.FastaFile(self.reference)

        bam = ps.AlignmentFile(self.bam_file)

        junctionmap = JunctionMap()

        # write junction reads information
        for chrom in CHROMS.keys():  # change CHROMS
            self.worker(bam, reference, chrom, self.quality, idn, junctionmap)

            # log.info(f'{chrom} FINSHED. NO PROBLEM FOUND')

        if self.output:
            junctionmap.write2file(self.output)  # write to file

        return junctionmap



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev
import io
import os
import sys
from shutil import rmtree

from setuptools import Command
from setuptools import find_packages
from setuptools import setup

# Package meta-data.
NAME = "ce-detector"
DESCRIPTION = "A program used to detect Cryptic Exon."
URL = "https://github.com/cauliyang/ce_detector/"
EMAIL = "li002252@umn.edu"
AUTHOR = "YangyangLi"
REQUIRES_PYTHON = ">=3.6.0"
VERSION = "0.1.12"

with open("HISTORY.rst") as history_file:
    history = history_file.read()

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest>=3",
]

# What packages are required for this module to be executed?

here = os.path.abspath(os.path.dirname(__file__))
try:
    with open(os.path.join(here, "requirements.txt"), encoding="utf-8") as f:
        REQUIRED = [line.strip() for line in f]
except FileNotFoundError:
    REQUIRED = [
        # 'requests', 'maya', 'records',
    ]

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

# here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print(f"\033[1m{s}\033[0m")

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system(f"{sys.executable} setup.py sdist bdist_wheel --universal")

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{}".format(about["__version__"]))
        os.system("git push --tags")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description + "\n\n" + history,
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],
    entry_points={
        "console_scripts": ["ce-detector=ce_detector.cli:cli"],
    },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    setup_requires=setup_requirements,  # add pytest
    test_suite="tests",
    tests_require=test_requirements,
    include_package_data=True,
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    # $ setup.py publish support.
    cmdclass={
        "upload": UploadCommand,
    },
    zip_safe=False,
)
